

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Accumulate_and_return Example Program using the 
**   LAPI_Amsend api call.  For a set of n tasks 0,1,...,n-1,
**   where n is an even number, all tasks are divided into
**   (src,tgt) buddy pairs (0,1), (2,3), etc.  For each pair,
**   an array of ints is transferred from src to tgt with a
**   LAPI_Amsend call.  The ints are then manipulated
**   and sent back to the src with a LAPI_Amsend call from
**   within the completion handler.  Since active messages are
**   used, all synchronization is done through handlers, and thus
**   LAPI counters are not needed.  The src does wait on its own
**   counter which is set in the src completion handler
**   The various data buffers are printed along the way to
**   show progress.
*/


/* Data structure for user header */
typedef struct {
    uint length;
    uint buddy;
} uhdr_t;

/* Data structure for info to pass to completion handler */
typedef struct {
    int *tmp_buf;
    uint length;
    uint buddy;
} compl_info_t;

compl_info_t compl_info;

/* stores initial value on src and tgt and final value on tgt */
int           data_buffer[ARRAYLEN];

/* global array of header handler addrs */
void   **hdr_hdl_list;

/* counter used on src (not a LAPI cntr) */
int wait_flag;


/*
**  src side handlers (invoked when tgt does return
**  Amsend call.
*/

/* completion handler on original src (as final tgt) */
void set_final_cntr(lapi_handle_t *handle, void *param)
{
    int i;     /* loop counter */
    compl_info_t *compl_ptr=(compl_info_t *)param;

    printf("Greetings from the src completion handler\n");
    for( i=0; i<compl_ptr->length; i++ ) {
        printf("final data_buf[%d]: %d\n", i,
                        data_buffer[i]);
    }
  
    /* set wait flag here to continue execution below */
    wait_flag++;

}

/* header_handler on src */
void *src_header_handler(lapi_handle_t *handle, void *uhdr,
                         uint uhdr_len, uint msg_len,
                         compl_hndlr_t **completion_handler,
                         void **user_info)
{
    uhdr_t *uhdr_ptr = (uhdr_t *) uhdr;  /* info from src */

    printf("Greetings from the src header handler...\n");

    *completion_handler = set_final_cntr;
    compl_info.length = uhdr_ptr->length;

    *user_info = (void *)&compl_info;

    /* LAPI will write data payload here */
    return data_buffer;
}

/*
**  tgt side handlers (invoked when src does return
**  Amsend call.
*/

/* completion handler on original target */
void do_accumulate(lapi_handle_t *handle, void *param)
{
    uhdr_t        uhdr;      /* pass info to HH on src */
    int i;                   /* loop counter */
    int buddy;               /* we send to orig. src */
    uint len;                /* pulled from param */
    int *tmp_buf;            /* pulled from param */

    /* grab info from compl_info */
    tmp_buf = (int *)(compl_info.tmp_buf);
    len     = (uint) (compl_info.length);
    buddy   = (uint) (compl_info.buddy);

    printf("Greetings from the tgt completion handler...\n");

    /* do the accumulation */
    for( i = 0; i < len; i++ ) {
        data_buffer[i] += tmp_buf[i];
    }

    /* setup for the return Amsend */
    uhdr.length = len;

    CHECK((LAPI_Amsend(*handle, buddy, 
                       (void *)(hdr_hdl_list[buddy]), &uhdr,
                        sizeof(uhdr_t), &(data_buffer[0]),
                        len*(sizeof(data_buffer[0])),
                        NULL, NULL, NULL)));

    /* done with tmp_buf */
    free(tmp_buf);
}

/* header handler on original tgt */
void *tgt_header_handler(lapi_handle_t *handle, void *uhdr,
                         uint uhdr_len, uint msg_len,
                         compl_hndlr_t **completion_handler,
                         void **user_info) {

    uhdr_t *uhdr_ptr;  /* for info from uhdr */
    int i;             /* loop counter */
    uint len;          /* stores length passed in user header */
    uint buddy;        /* pulled for uhdr */ 
    int *tmp_buf;      /* intermediate buffer for accumulate */

    printf("Greetings from the tgt header handler...\n");
    
    /* grab the length passed in the user header */
    uhdr_ptr = (uhdr_t *) uhdr;
    len = uhdr_ptr->length;
    buddy = uhdr_ptr->buddy;

    /* 
    ** malloc to buffer then pass back to LAPI for use in
    ** completion handler.  completion handler will free.
    */ 
    tmp_buf = (int *) malloc(len*sizeof(int));

    /* store some arbitrary values in tmp_buf */
    for (i = 0; i < len; i++ ) {
        tmp_buf[i] = (len) - (2*i);
    }   

    /* set completion handler pointer */
    *completion_handler = do_accumulate;

    /* set params to send to completion handler */
    compl_info.tmp_buf = tmp_buf;
    compl_info.length = uhdr_ptr->length;
    compl_info.buddy = buddy;

    return data_buffer;
}

int main( int argc, char **argv )
{

lapi_handle_t handle;                      /* LAPI handle */
lapi_info_t   info;                        /* Info to pass to
                                              LAPI_Init */
int           task_id;                     /* Our LAPI Task id */
int           num_tasks;                   /* Total number of tasks
 */
void          *buf_addr_list[NUM_TASKS];   /* Table of remote data
                                              buf addrs. */
int           i;                           /* Loop counter */
int           buddy;                       /* Our communication
                                              partner */

    /*
    ** Clear the struct.  Not passing any info to init through this
    ** struct.
    */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks < 2) || ((num_tasks % 2) != 0) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }

    /* setup for header handler addresses */
    hdr_hdl_list = (void **) malloc(num_tasks*sizeof(void *));

    /*
    ** up to this point, all instructions have executed on all 
    ** tasks. we now begin differentiating tasks.
    */
    if ((task_id %2) == 0) {   /* message source */
        uhdr_t        uhdr;

    /* setup buddy pairs as (0,1), (2,3), etc. */
        buddy = task_id+1;

    /*
    ** collective calls.  in this case, we need to specify
    ** different header handlers for the src and tgt tasks,
    ** thus the Address_init call is done within the separate 
    ** code for each task.
    */
        CHECK((LAPI_Address_init(handle,(void *)&src_header_handler,hdr_hdl_list)));

    /* for our own sync. */
        wait_flag = 0;

    /* initialize data buffer */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = i;
            printf("original data_buffer[%d]: %d\n", i, data_buffer[i]);
        }

        /* sync before starting */
        CHECK((LAPI_Gfence(handle)));

        /* initialize our header */
        uhdr.length = ARRAYLEN;

        /* tgt will use in return Amsend call. */
        uhdr.buddy = task_id;  
    /*
    ** execute the data transfer to our buddy task.  send ARRAYLEN
    ** ints, starting with data_buffer[0].  Note that no user
    ** header is sent, and no counters are used.
    ** args:  handle                 -- the LAPI handle
    **        buddy                  -- the target task id
    **        hdr_hdl_list[buddy]    -- the header handler that
    **                                  will execute on the target
    **        uhdr                   -- the user header
    **        sizeof(uhdr_t)         -- the user header length
    **        &(data_buffer[0])      -- the starting address of
    **                                  data to transfer
    **        ARRAYLEN*sizeof(int)   -- the length of data to
    **                                  transfer
    **        NULL, NULL, NULL       -- target, origin and 
    **                                  completion counters
    */

        CHECK((LAPI_Amsend(handle, buddy,
                         (void *)hdr_hdl_list[buddy], &uhdr, 
                         sizeof(uhdr_t),&(data_buffer[0]),
                         ARRAYLEN*(sizeof(data_buffer[0])),
                         NULL, NULL, NULL)));

        /* wait for src compl hndlr to signal that we're done */
        while (!wait_flag) {
            sleep(1);
        }
        


    } else {             /* receiver */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = 2*(-i);
        }

    /*
    ** collective calls.  in this case, we need to specify
    ** different header handlers for the src and tgt tasks,
    ** thus the Address_init call is done within the separate
    ** code for each task.
    */
    CHECK((LAPI_Address_init(handle,(void *)&tgt_header_handler,hdr_hdl_list)));

        /* sync with src */
        CHECK((LAPI_Gfence(handle)));
    }

    /* remaining code will execute on all tasks */ 

    /* final sync before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(hdr_hdl_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


