

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Accumulate_and_return Example Program using the LAPI_Xfer
**   interface to the Put api call.  For a set of n tasks
**   0,1,...,n-1, where n is an even number, all tasks are divided
**   into (src,tgt) buddy pairs (0,1), (2,3), etc.  For each pair,
**   an array of ints is transferred from src to tgt with a
**   LAPI_Xfer call.  The ints are then manipulated and sent back
**   to the src with a regular LAPI_Put call.  All tasks wait on
**   target counters to signal completion of message delivery.
**   The various data buffers are printed along the way to show
**   progress.
*/


/* stores initial value on src and tgt and final value on src */
int           data_buffer[ARRAYLEN];

lapi_long_t   *data_buffer_list;  /* remote data buffer addrs */
lapi_long_t   *tgt_cntr_list;     /* tgt cntr is remote addr. */

/* 
** Task 1 will wait in tgt_cntr for completion of first data xfer.
** Task 0 will wait on tgt_cntr for completion of data return.
*/
lapi_cntr_t   tgt_cntr;   /* updates on tgt at msg completion */

int main( int argc, char **argv )
{

lapi_handle_t handle;        /* LAPI handle */
lapi_info_t   info;          /* Info to pass to LAPI_Init */
int           task_id;       /* Our LAPI Task id */
int           num_tasks;     /* Total number of tasks */
int           i,j;           /* Loop counters */
int           buddy;         /* Our communication partner */
lapi_xfer_t   xfer_info;     /* for xfer call */
int           val;           /* Needed for waitcntr call on source
*/

    /* Not passing any info to init through this struct. */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks < 2) || ((num_tasks % 2) != 0) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }

    /* each task executes on tgt_cntr within its own data space */
    CHECK((LAPI_Setcntr(handle,&tgt_cntr,0)));

    /* allocate space for remote addrs */
    data_buffer_list = (lapi_long_t *) 
                     malloc(num_tasks*sizeof(lapi_long_t));
    tgt_cntr_list = (lapi_long_t *)
                     malloc(num_tasks*sizeof(lapi_long_t));

    /*
    ** collective calls.  each task stores its own addr plus
    ** remote addrs
    */
    CHECK((LAPI_Address_init64(handle,(lapi_long_t)
                                     &tgt_cntr,tgt_cntr_list)));
    CHECK((LAPI_Address_init64(handle,(lapi_long_t)
                                data_buffer,data_buffer_list)));

    if ((task_id %2) == 0) {   /* sender */
        buddy = task_id + 1;
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = i;
        }

        for (i = 0; i < ARRAYLEN; i++ ) {
            printf("original data_buffer[%d]: %d\n", i, data_buffer[i]);
        }

        /* sync before data transfer */
        CHECK((LAPI_Gfence(handle)));


        /* Set up xfer structure */
        bzero(&xfer_info, sizeof(lapi_xfer_t));
        xfer_info.Put.Xfer_type = LAPI_PUT_XFER;

        /* send data to buddy */
        xfer_info.Put.tgt = buddy;

        /* data size */
        xfer_info.Put.len = (ulong)(ARRAYLEN*sizeof(int));

        /* remote base address for writing data */
        xfer_info.Put.tgt_addr = (lapi_long_t)
                                 (data_buffer_list[buddy]);
        
        /* local base address for reading data */
        xfer_info.Put.org_addr = (void *) (&(data_buffer[0]));

        /* target counter */
        xfer_info.Put.tgt_cntr = (lapi_long_t)
                                 (tgt_cntr_list[buddy]);

        /* not using origin or completion counters */
        xfer_info.Put.org_cntr = NULL;
        xfer_info.Put.cmpl_cntr = NULL;

        /* do the xfer call */
        CHECK((LAPI_Xfer(handle, &xfer_info)));

        /* 
        ** now wait for tgt to return data to us by waiting
        ** on local tgt_cntr to increment
        */
        CHECK((LAPI_Waitcntr(handle, &tgt_cntr,
                                                1, &val)));

        for (i = 0; i < ARRAYLEN; i++ ) {
            printf("final data_buffer[%d]: %d\n", i,
                                            data_buffer[i]);
        }


    } else {             /* receiver */
        int tmp_buf[ARRAYLEN];  /* buffer to add to orig */  

        buddy = task_id -1;
         
        for( j = 0; j < ARRAYLEN; j++ ) {
            tmp_buf[j] = ARRAYLEN - (2*j);
        }

        /* match src's sync */       
        CHECK((LAPI_Gfence(handle)));

        /* wait for data to arrive from src */
        CHECK((LAPI_Waitcntr(handle, &tgt_cntr,1,
                                         &val)));

        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] += tmp_buf[i];
        }

        /* return buffer to src with */
        CHECK((LAPI_Put(handle, buddy, (ulong)
                          (ARRAYLEN*sizeof(int)), 
                          (void *) (data_buffer_list[buddy]),
                          (void *) (&(data_buffer[0])), 
                          (lapi_cntr_t *)tgt_cntr_list[buddy],
                          NULL, NULL)));  
    }

    /* all tasks will execute remaining code */ 

    /* final sync before term */ 
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(data_buffer_list);
    free(tgt_cntr_list);

    /* terminate the LAPI handle */ 
    CHECK((LAPI_Term(handle)));

    return 0; 
}


