

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
**   Example Program to illustrate the various approaches to
**   initializing LAPI.  The use of various init parameters and LAPI's ability 
**   to re-initialize a handle after termination are illustrated.
**   For a set of n tasks 0, 1,..., n-1, where n is an even number, a 
**   sequence of LAPI initialzation, Gfence, and termination is performed.  
**
**   The sequence of calls is as follows:
**   - A single LAPI handle is inited.
**   - Simple qenv data for this handle is shown.
**   - Gfence is performed to synchronize among all tasks.
**   - The handle is terminated.
**   - The handle is re-initialized. 
**   - Gfence is performed to synchronize among all tasks.
**   - The handle is terminated again.
**
**   Note that the use of an error handler is also illustrated.
**   Each handle registers its own error handler through the 
**   lapi_info_t struct.
*/
 
/*
**  The handler to invoke on native LAPI handles on error condition.
*/
void user_lapi_err_hndlr(lapi_handle_t *hndl, int *error_code,
                         lapi_err_t *err_type, int *task_id,
                         int *src)
{
    printf("Greetings from the LAPI error handler.\n");
}

int main( int argc, char **argv )
{
    lapi_handle_t lapi_handle;   /* LAPI handle */
    lapi_info_t   lapi_info;     /* Info to pass to LAPI_Init */
    int           task_id;       /* Our LAPI Task id */
    int           s_task_id;     /* shared handle */
    int           num_tasks;     /* number of tasks in job */
    int           i,j;           /* Loop counters */

    /* Clear the struct (values in future fields will cause error). */
    bzero(&lapi_info, sizeof(lapi_info_t));

    /* Set Error handler */
    lapi_info.err_hndlr = user_lapi_err_hndlr;

    /* Initialize the LAPI handle */
    printf("Initializing native LAPI Handle.\n");
    CHECK((LAPI_Init(&lapi_handle, &lapi_info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(lapi_handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(lapi_handle, NUM_TASKS, &num_tasks)));
    printf("LAPI handle shows task_id %d and %d tasks.\n", task_id, num_tasks);

    /* all tasks will execute this before term */
    printf("Gfence on LAPI handle\n");
    CHECK((LAPI_Gfence(lapi_handle)));

    /* terminate the LAPI handle */
    printf("Terminate the LAPI handle\n");
    CHECK((LAPI_Term(lapi_handle)));

    /* re-Initialize the LAPI handle */
    CHECK((LAPI_Init(&lapi_handle, &lapi_info)));
    printf("re-Initializing native LAPI Handle.\n");

    /* do an operation to confirm success of re-init */
    printf("Gfence on re-init LAPI handle\n");
    CHECK((LAPI_Gfence(lapi_handle)));

    /* final termination of the LAPI handle */
    printf("Terminate the LAPI handle\n");
    CHECK((LAPI_Term(lapi_handle)));

    return 0;
}


