

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* constant for array lengths */
#define ARRAYLEN  10

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Example Program to illustrate the use of the LAPI_Get api call.
**   For a set of n tasks 0,1,...,n-1, where n is an even number, all
**   tasks are divided into (src,tgt) buddy pairs (0,1), (2,3), etc.
**   For each pair, an array of ints is transferred from tgt to src
**   with a single LAPI_Get call.  The origin counter is used for
**   synchronization.  Once the message completes, the origin shows
**   the result by printing the contents of the data buffer.
*/


/* stores initial value on tgt and final value on src */
int           data_buffer[ARRAYLEN];

/* stores list of remote buffer addrs */
void          **data_buffer_list;

/* updates on src at msg completion */
lapi_cntr_t   origin_cntr;


int main( int argc, char **argv )
{

lapi_handle_t handle;        /* LAPI handle */
lapi_info_t   info;          /* Info to pass to LAPI_Init */
int           task_id;       /* Our LAPI Task id */
int           num_tasks;     /* Total number of tasks */
int           i,j;           /* Loop counters */
int           buddy;         /* Our communication partner */
int           val;           /* Needed for waitcntr call on source */

    /* Not passing any info to init through this struct. */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks < 2) || ((num_tasks % 2) != 0) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }

    /* initialize the table buffer for the data addrs. */
    data_buffer_list = (void **) malloc(num_tasks*sizeof(void *));

    /* collective call.  each task stores its own addr plus remote addrs */
    CHECK((LAPI_Address_init(handle,(void *)data_buffer,data_buffer_list)));

    /*
    ** up to this point, all instructions have executed on all tasks.
    ** we now begin differentiating tasks.
    */
    if ((task_id %2) == 0) {   /* message source */

        /* setup buddy pairs as (0,1), (2,3), etc. */
        buddy = task_id + 1;

        /* initialize origin counter */
        CHECK((LAPI_Setcntr(handle,&origin_cntr,0)));

        /* collective call.  sync before starting data transfer */
        CHECK((LAPI_Gfence(handle)));

    /*
    ** execute the data transfer from our buddy task. Retrieve ARRAYLEN
    ** ints, starting with data_buffer[0].  Note that the origin counter
    ** is used to signal message completion at the source.
    ** args:  handle                  -- the LAPI handle
    **        buddy                   -- the target task id
    **        ARRAYLEN*sizeof(int)    -- the length of data to transfer
    **        data_buffer_list[buddy] -- remote addr for writing data
    **        &(data_buffer[0])       -- the starting address of data
    **                                   to transfer
    **        NULL                    -- target counter
    **        &origin_cntr            -- origin counter (will update
    **                                   when data completes
    **                                   arrival at src)
    */
        CHECK((LAPI_Get(handle, buddy, 
                           (ulong) (ARRAYLEN*sizeof(int)), 
                           (void *) (data_buffer_list[buddy]),
                           (void *) (&(data_buffer[0])), 
                           NULL, &origin_cntr)));  

        /* wait for data to arrive from tgt */
        CHECK((LAPI_Waitcntr(handle, &origin_cntr,1, &val)));
        printf("Received data from buddy\n");
        for (i = 0; i < ARRAYLEN; i++ ) {
            printf("data_buffer[%d]: %d\n", i, data_buffer[i]);
        }

    } else {             /* receiver */

        /* buddy = task_id - 1; */

        /* 
        ** initialize the data buffer (tgt inits the buffer for the
        ** Get call)
        */
        for (i = 0; i < ARRAYLEN; i++ ) {
            data_buffer[i] = i*(task_id);
        }

        /* to match the src's sync before data transfer */
        CHECK((LAPI_Gfence(handle)));

    }

    /* all tasks will execute this before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(data_buffer_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


