

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <lapi.h>

#define VECLEN         15
#define NUM_VECS       10
#define ARRAYLEN       10

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];



/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CHECK(func_and_args)                                      \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
                                                                  \
}
 


/*
** Example Program to illustrate the use of the LAPI_Getv api call.
** For a set of n tasks 0,1,...,n-1, where n is an even number, all
** tasks are divided into (src,tgt) buddy pairs (0,1), (2,3), etc.
** For each pair, a set of LAPI vectors is transferred from tgt to src
** with a single LAPI_Getv call.  Each vector contains a single int.
**   
** The origin counter is used for synchronization.  Once the message
** completes, the src shows the result by printing the contents of
** the data buffer.
*/

/* stores initial value on tgt */
int           data_buffer[NUM_VECS];

/* store list of remote buffer addrs */
void          **data_buffer_list;

/* updates on src at msg completion */
lapi_cntr_t   origin_cntr;



int main( int argc, char **argv )
{

lapi_handle_t handle;           /* LAPI handle */
lapi_info_t   info;             /* Info to pass to LAPI_Init */
int           task_id;          /* Our LAPI Task id */
int           num_tasks;        /* Total number of tasks */
int           i,j;              /* Loop counters */
int           buddy;            /* Our communication partner */
int           val;              /* Needed for waitcntr call on source */
lapi_vec_t *org_vec, *tgt_vec;  /* setup data transfer on both sides */

    /* Not passing any info to init through this struct. */
    bzero(&info, sizeof(lapi_info_t));

    /* Initialize the LAPI handle */
    CHECK((LAPI_Init(&handle, &info)));

    /* Query LAPI for our task id */
    CHECK((LAPI_Qenv(handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CHECK((LAPI_Qenv(handle, NUM_TASKS, &num_tasks)));

    /* This example only supports even numbers of tasks */
    if ( (num_tasks < 2) || ((num_tasks % 2) != 0) ) {
        fprintf(stderr,"ERROR: this example requires an even number of tasks, but has been invoked with %d\n", num_tasks);
        exit(1);
    }

    /* allocate the table buffer for the data addrs. */
    data_buffer_list = (void **) malloc(num_tasks*sizeof(void *));
    /*
    ** collective call.  each task stores its own addr plus remote
    ** addrs
    */
    CHECK((LAPI_Address_init(handle,(void *)data_buffer,
                                         data_buffer_list)));

    /*
    ** up to this point, all instructions have executed on all tasks.
    ** we now begin differentiating tasks.
    */
    if ((task_id %2) == 0) {   /* message source */

        /* setup buddy pairs as (0,1), (2,3), etc. */
        buddy = task_id + 1;
 
        /*
        ** Setup the vectors for a basic vector transfer.  Both vectors
        ** must have the same type (LAPI_GEN_IOVECTOR).  As a
        ** GEN_IOVECTOR transfer,the transfer vectors must have the
        ** same structure, that is, they contain the same number of
        ** subvectors and each subvector on the tgt is of the same
        ** length as the corresponding subvector on the src.  In
        ** this instance, each subvector contains a single int.
        ** 
        ** The setup of BOTH the org_vec and tgt_vec is done by the src.
        ** In particular, the src allocates storage to the pointers for
        ** both structures, including storage for the array of data i
        ** buffers and the array of data lengths.  For both the org_vec
        ** and the tgt_vec, the array of data lengths is initialized for
        ** a single int (sizeof (int)) to be transferred.  The array of
        ** data buffers on the tgt_vec is initialized with the remote
        ** addresses stored in the data_buffer_list structure that
        ** resulted from the LAPI_Address_init call above.  Each address
        ** refers to an element of the data_buffer structure in the tgt
        ** address space.  These addresses refer to the data to be
        ** transferred by the Getv call.  The array of data buffers
        ** in the org_vec do not need to be initialized before the Getv
        ** call (other than to have sufficient storage allocated).  These
        ** buffers will contain the RESULT of the Getv call.  
        */
 
        /* allocate memory for each structure */
        org_vec = (lapi_vec_t *)malloc(sizeof(lapi_vec_t));
        tgt_vec = (lapi_vec_t *)malloc(sizeof(lapi_vec_t));


        /*
        ** Setup org_vec.
        */
 
        /* Set vector type and number of subvectors */
        org_vec->vec_type = LAPI_GEN_IOVECTOR;
        org_vec->num_vecs  = NUM_VECS;

        /* allocate storage for arrays of data buffers and lengths */ 
        org_vec->info = (void **) malloc(NUM_VECS * sizeof(void *));
        org_vec->len = (unsigned long *) malloc(NUM_VECS * sizeof(ulong));

        /* initialize lengths array and allocate space for data buffers */
        for ( j = 0; j < NUM_VECS; j++ ) {
            org_vec->len[j] = (unsigned long)sizeof(int);
            org_vec->info[j] = (void *)malloc(sizeof(int));
        }


        /*
        ** Setup tgt_vec.
        */
 
        /* Set vector type and number of subvectors */
        tgt_vec->vec_type = LAPI_GEN_IOVECTOR;
        tgt_vec->num_vecs = NUM_VECS;

        /* allocate storage for arrays of data buffers and lengths */
        tgt_vec->info = (void **) malloc(NUM_VECS * sizeof(void *));
        tgt_vec->len = (unsigned long *) malloc(NUM_VECS * sizeof(ulong));

        /*
        ** Initialize array of lengths and assign remote addrs to
        ** data array.  Remote addrs are pulled from data_buffer_list
        ** that was filled in LAPI_Address_init call above.  We offset
        ** each successive address by sizeof(int) to point to the next 
        ** element in the data_buffer array.  
        */
        for ( j = 0; j < NUM_VECS; j++ ) {
            tgt_vec->info[j] = (void *)(data_buffer_list[buddy]
                               + (j*sizeof(int)));
            tgt_vec->len[j] = (unsigned long)sizeof(int);
        }

        /* initialize origin counter */
        CHECK((LAPI_Setcntr(handle,&origin_cntr,0)));

        /* collective call.  sync before starting data transfer */
        CHECK((LAPI_Gfence(handle)));

    /*
    ** Execute the data transfer from our buddy task.  Receive the
    ** vectors using the addresses and lengths specified in tgt_vec
    ** (NUM_VECS vectors of one int each.)  Remote addresses point to
    ** individual elements of the data_buffer array in the tgt address
    ** space.  Note that the origin counter is used to signify message
    ** completion on the src side.
    ** args:  handle       -- the LAPI handle
    **        buddy        -- the target task id
    **        tgt_vec      -- a pointer to the target vector description
    **        org_vec      -- a pointer to the origin vector description
    **        NULL         -- target counter
    **        &origin_cntr -- origin counter (will update when data 
    **                                        completes arrival at src)
    */
        CHECK((LAPI_Getv(handle, buddy, tgt_vec, org_vec, 
                           NULL, &origin_cntr)));  

        /* wait for data to arrive from src */
        CHECK((LAPI_Waitcntr(handle, &origin_cntr,1, &val)));

        printf("Received data from buddy %d\n",buddy);
        for( i=0; i < NUM_VECS; i++ ) {
            printf("org_vec->info[%d]: %d\n",i,*(int *)(org_vec->info[i]));
        }

        /* free all vector structs */
        for( i=0; i < NUM_VECS; i++ ) {
           free(org_vec->info[i]);
        }
        free(org_vec->info);
        free(org_vec->len);
        free(org_vec);
        free(tgt_vec->info);
        free(tgt_vec->len);
        free(tgt_vec);

    } else {             /* receiver */
        /* buddy = task_id -1 */

        /*
        ** initialize the data buffer (tgt inits the buffer for the
        ** Get call)
        */
        for( i=0; i < NUM_VECS; i++ ) {
            data_buffer[i]= i;
        }

        /* to match the src's sync before data transfer */
        CHECK((LAPI_Gfence(handle)));

    }

    /* all tasks will execute this before term */
    CHECK((LAPI_Gfence(handle)));

    /* cleanup */
    free(data_buffer_list);

    /* terminate the LAPI handle */
    CHECK((LAPI_Term(handle)));

    return 0;
}


