
/*
**   Example Program to illustrate initialization of LAPI for IP
**   communication in standalone mode, that is without using IBM's
**   Parallel Environment (PE) for AIX.  In the absence of PE, the user
**   must provide a mechanism for distributing all-pairs IP address and
**   port information for each task.  For each task wishing to engage
**   in IP communication, LAPI opens two sockets, one for reading and one
**   for writing using connectionless (i.e. UDP) communication.  Each
**   task must know the IP address and port number on which the other
**   tasks are listening for reading in order to write to that socket.
**
**   When LAPI is used with PE, PE is responsible for telling LAPI the
**   IP address of the communication device selected by the user.  LAPI
**   then opens the socket connections (allowing the operating system
**   to pick the port number) and passes this information back to PE.
**   PE then maintains a repository of IP address and port
**   information for all tasks which LAPI can query when needed for 
**   communication with another task.
**
**   When not using PE, the user must provide a means for distributing
**   this information to the tasks.  LAPI supports two approaches by
**   which the user can distribute the required information.  At 
**   the initialization of a LAPI handle, the user may pass in either
**   a list of IP addresses and ports for all tasks in the job 
**   (user list), or the user may pass a pointer to a callback function
**   (user handler).  The (lapi_extend_t *) add_info member of the 
**   lapi_info_t structure is used to pass the required information.  In
**   the case of a user list, the user passes a pointer to a filled_in
**   array of type lapi_udp_storage, with one entry for each task.  In the
**   case of a user handler, the user passes a pointer to a function
**   pointer of type udp_init_hndlr.  LAPI expects this routine to
**   return (by reference) a pointer to a filled-in array of lapi_udp_storage,
**   with one entry for each task in the job.  See 
**   /usr/include/lapi.h for definitions of the lapi_info_t, 
**   lapi_extend_t, udp_init_hndlr and lapi_udp_storage types, as well as other
**   relevant data structure definitions.  
**
**   In either the user list or user handler case, LAPI uses the
**   filled-in lapi_udp_storage array to 1) open the required sockets on each
**   task and 2) setup an internal table of IP information for all of the
**   other tasks.  Since LAPI can only call the user handler (or parse
**   the user list) once, all information for all tasks must be 
**   provided.  One result of this fact is that a unique IP port number
**   must be provided by the user for each task.  LAPI cannot allow the
**   operating system to choose an available port at the time of socket
**   binding since there would be no way to communicate the chosen port
**   number to the other tasks in the job.  Care must therefore be
**   exercised to not assign the same port number to two tasks running
**   on the same node.
**   
**   This sample illustrates the use of both types of standalone IP
**   initialization for LAPI.  The method is chosen by command-line
**   argument when the sample is run.  If the user runs with '-hndlr'
**   then the handler method is used.  If the user runs with '-list'
**   then the list method is used.  Any other arguments will cause a
**   usage message to print.  Note that in standalone mode, each task
**   must be fired off individually.  If the user program (e.g. this
**   src) is written and built correctly, it may be possible to execute
**   tasks mixed between the handler and list methods (i.e. some use
**   handler, some use list) as long as the correct information is
**   disseminated. 
**    
**
**   For a set of n tasks 0,1,...,n-1, where n is an even number, an 
**   individual task is invoked each time the user executes the 
**   executable.  For each execution, LAPI initializes a handle
**   (including all of the steps described above for retrieving 
**   IP information and initializing sockets).  Then each task waits
**   in Gfence before terminating.  The Gfence is done in this example
**   for two reasons:  1) Gfence requires collective communication and
**   therefore demonstrates that a successful link has been established.
**   2) Since the tasks must be started by hand, there is a long lag
**   before all tasks initialize.  Gfence ensures full initialization
**   before any terminations.
**
**   To use this sample with a user list (invoked with -list):
**   The ip and port information must be included in this src code and
**   must thus be known when the sample is built.  For ease of alteration,
**   this sample was designed for all even-numbered tasks to be assigned
**   to a single host and all odd-numbered tasks to be assigned to
**   another single host.  The src file must be edited to correctly
**   define the EVEN_IP_STRING and ODD_IP_STRING constants.  These
**   constants are used in the setup_addr_list routine to create the user
**   list that is passed to LAPI.  The BASE_PORT constant must also be
**   properly defined and is used as follows:
**   For a given task n, BASE_PORT+n will be passed to LAPI as the
**   chosen port for UDP reads.  Straight incrementing is done (rather
**   than sharing port numbers between odd and even tasks) as a safety
**   measure in case the odd host and even host happen to be the same.
**   It is recommended that the port be chosen in the ephemeral port range
**   (typically, 49152 to 65535), preferably closer to the higher end of the
**   range.
**
**   
**   To use this example with a handler (invoked with -hndlr):
**   The information is read from a flat file and can thus be altered
**   at runtime (i.e. without recompiling the example).  The IP_FILE
**   envronment variable must contain the name of a readable file containing
**   the address and port information.  This file is read and the
**   information assigned.  The format of the IP_LIST file is:
**
**   dotted_quad_0                # e.g., 2001:250:5802:2:210:dcff:fe29:af32
**   port_0                       # e.g., 62555
**   dotted_quad_1                # e.g., 2001:250:5802:2:210:dcff:fe29:af33
**   port_1                       # e.g., 62555 (note, same port as task 0 
**                                               but different IP address)
**   ...
**   dotted_quad_num_tasks-1      # e.g., 2001:250:5802:2:210:dcff:fe29:af32
**   port_num_tasks-1             # e.g., 62565 (note, same IP address as task 0
**                                #              but different port number)
**                                # note: the ip addresses also can be hostname
**                                                                             
**   where dotted_quad_i is the dotted quad formatted ip address of
**   the device on which task i will read incoming data and port_i is
**   the port on which task i will listen. The examples above show that
**   IP addresses and port numbers can be re-used, but any given task must
**   be given a unique _combination_ of IP address and port over which it
**   can communicate.
**
**   Note also that the set of runtime environment variables must be
**   setup differently for standalone mode and the each task must be
**   fired off separately by the user.  In addition to the src code
**   changes documented here, see the UDP sections of the
**   README.LAPI.STANDALONE file in this directory for detailed
**   instructions on building and running this example.  
*/


/* General definitions for use by LAPI samples programs */ 
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>

#include "../include/lapi.h"

/* for any error messages returned by LAPI */
char err_msg_buf[LAPI_MAX_ERR_STRING];

/* 
** macro to check return code of function calls. keeps return
** code checking logic from needing to be in main logic 
*/
#define CALL_AND_CHECK_RC(func_and_args)                          \
{                                                                 \
    int rc;                                                       \
    if ((rc = (func_and_args)) != LAPI_SUCCESS) {                 \
        LAPI_Msg_string(rc, err_msg_buf);                         \
        fprintf(stderr,                                           \
                "LAPI returns error message: %s, rc = %d\n",      \
                 err_msg_buf, rc);                                \
        exit(1);                                                  \
    }                                                             \
}
 


/***** Begin function definitions *****/

/* print a usage message if something doesn't jive with the */
/* command line args. */
void usage()
{
    printf("usage: Init.standalone.udp [-hndlr|-list]\n");
    printf("\t-hndlr does IP initialization by callback\n");
    printf("\t-list does IP initialization by list\n");
}    

/* Setup the initial environment, reading command line args and */
/* number of tasks from env var */
int do_setup(int argc, char** argv, int *hndlr, int* num_tasks)
{
    char *user_arg = NULL;   /* for cmd line argument */
    char *mp_procs = NULL;   /* for MP_PROCS */

    if ((argc < 2) || (argc > 3)) {
        usage();
        return(1);
    }          

    /* above check ensures exactly one arg in argv */
    user_arg = argv[1]; 

    /* check for user's choice of list or hndlr */
    if (!strncasecmp(user_arg,"-list",5)) { /* list */ 
        *hndlr = 0;
    } else if (!strncasecmp(user_arg,"-hndlr",6)) { /* hndlr */
        *hndlr = 1;
    } else {
        fprintf(stderr,"Error: option '%s' unknown.\n", user_arg);
        usage();
        return 1;
    }   

    /*
    ** query number of tasks from the user environment.  Note that when
    ** running in standalone mode, this information must be known
    ** before calling LAPI_Init and thus cannot be queried from LAPI
    ** as is normally done.
    */ 
    mp_procs = getenv("MP_PROCS");
    *num_tasks = atoi(mp_procs);
    if ( *num_tasks <= 0 ) {
        fprintf(stderr,"Error: MP_PROCS is %s, " 
                "should be the number of tasks.", mp_procs);
        return 1;
    }

    return LAPI_SUCCESS;    
}
   
/*
** Following #defines are for use with the -list option only. They
** are ignored with the -hndlr option.
*/ 
#define EVEN_IP_STRING "2001:250:5802:2:210:dcff:fe29:af32"   /* change to addr for even tasks */
#define ODD_IP_STRING  "2001:250:5802:2:210:dcff:fe29:af33"   /* change to addr for odd tasks */
#define BASE_PORT      62555           /* change this to the base port number */ 

/* 
** Setup the user list using the values defined in the above
** constants.  addr_list will be returned by reference.
*/

int inet_addr_ext(char *str_addr, lapi_udp_storage *addr)
{
	struct addrinfo *host;
	char hostnum[NI_MAXHOST];
	char service[NI_MAXSERV] = "0";

	getaddrinfo(str_addr, service, NULL, &host);
	if (host) {
		getnameinfo(host->ai_addr, host->ai_addrlen, hostnum, sizeof(hostnum), NULL, 0, NI_NUMERICHOST);
		addr->af = host->ai_family;
		inet_pton(addr->af, hostnum, (void *)&addr->ip_addr);
	}
	freeaddrinfo(host);

    return 0;
}

int setup_addr_list(lapi_udp_storage *addr_list, int num_tasks)
{
    int i;               /* loop counter */

    printf("Using user list.\n");

    /* set ip, port info array */
    for( i = 0; i < num_tasks; i++ ) {
        if ( i % 2 ) { /* odd */
            inet_addr_ext(ODD_IP_STRING, &addr_list[i]); 
        } else {
            inet_addr_ext(EVEN_IP_STRING, &addr_list[i]); 
        }
        addr_list[i].port_no = htons(BASE_PORT+i);            
        printf( "Assigning addr_list[%d].ip_addr:%lu\n",
                        i, addr_list[i].ip_addr);
        printf( "Assigning addr_list[%d].port_no:%u\n",
                        i, addr_list[i].port_no);
    }
    return LAPI_SUCCESS;
}

#define MAX_STRING_LEN 80

/*
** Read and parse information from the file specified in the "IP_FILE"
** env var.  Initialize the addr_list data structure according to the 
** format that LAPI expects.
*/  
int udp_hndlr(lapi_handle_t *hndl, lapi_udp_storage *local_addr,
              lapi_udp_storage *addr_list, lapi_udpinfo_t *info) 
{
    int i;                              /* loop counter */
    int num_tasks = info->num_tasks;    /* number of tasks */
    unsigned port_as_unsigned;          /* unsigned port */
    char *ip_filename;                  /* name fo file from IP_FILE env */
    FILE *ip_file;                      /* FILE pointer for reading */
    char infile_string[MAX_STRING_LEN]; /* string buffer to read into */

    printf("Greetings from the UDP HANDLER\n");

    /* setup FILE ptr */
    ip_filename = getenv("IP_FILE");
    if (!ip_filename) {
        fprintf(stderr,"Error, UDP Handler requires IP_FILE environment "
                       "variable to be set.\n");
        return (1);
    }
    ip_file = fopen(ip_filename, "r");
    if ( !ip_file ) {
        fprintf(stderr,"Error, opening %s for reading, errno:%d.\n",
                ip_filename, errno);
        return(1);
    }

    /*
    ** read and parse info from file.  store in correct format in
    ** return param.
    */
    for ( i = 0; i < num_tasks; i++ ) {
        fgets(infile_string, MAX_STRING_LEN, ip_file);
        printf("Setting ip %d to %s\n", i, infile_string);
		infile_string[strlen(infile_string)-1] = '\0';
        inet_addr_ext(infile_string, &addr_list[i]);

        fgets(infile_string, MAX_STRING_LEN, ip_file);
        printf("Read port %d: %s\n", i, infile_string);
        port_as_unsigned = (unsigned)(atoi(infile_string));
        printf("Setting Port %d to %u\n", i, port_as_unsigned);
        addr_list[i].port_no = htons(port_as_unsigned);
    }
    return LAPI_SUCCESS;
}


int main( int argc, char **argv )
{
    int           rc;            /* Return code from func. calls */
    int           use_hndlr = 0; /* hndlr or list? */
    lapi_handle_t lapi_handle;   /* LAPI handle */
    lapi_info_t   lapi_info;     /* Info to pass to LAPI_Init */
    lapi_extend_t extend_info;   /* Structure to hold IP addresses and ports */
    int           task_id;       /* Our LAPI Task id */
    int           num_tasks;     /* number of tasks in job */
    int           i, j;          /* Loop counters */
    lapi_udp_storage    *udp_info = NULL;     /* List of ip, port info to pass to LAPI */


    if ( (rc = do_setup(argc, argv, &use_hndlr, &num_tasks)) != 0 ) {
        fprintf(stderr,"Error during setup\n");
        exit(rc);
    }

    /* Clear the struct (values in future fields will cause error). */
    bzero(&lapi_info, sizeof(lapi_info_t));
    bzero(&extend_info, sizeof(lapi_extend_t));

    /* Assign the address of the lapi_extend_t to pass to LAPI */
    lapi_info.add_info = (lapi_extend_t *)&extend_info;

    if ( !use_hndlr) {
        udp_info = (lapi_udp_storage *) malloc(num_tasks*sizeof(lapi_udp_storage));
        if ( (rc = setup_addr_list(udp_info, num_tasks)) != 0 ) {
            fprintf(stderr,"Error during setup_addr_list\n");
            exit(rc);
        }

        /* assign list to info struct */
        lapi_info.add_info->add_udp_ext_addrs = udp_info; 
        lapi_info.add_info->num_udp_addr = num_tasks;

        /* clear handler so not used */
        lapi_info.add_info->udp_ext_hndlr = 0;

    } else {  /* setup for use of hndlr */    
        lapi_info.add_info->udp_ext_hndlr = udp_hndlr;
        lapi_info.add_info->add_udp_ext_addrs = 0;
    }
    printf("lapi_info.add_info->add_udp_ext_addrs:0x%x\n",
           lapi_info.add_info->add_udp_ext_addrs);
    printf("lapi_info.add_info->udp_ext_hndlr:0x%x\n",
           lapi_info.add_info->udp_ext_hndlr);

    /* Initialize the LAPI handle */
    printf("Initializing native LAPI Handle, add_info:0x%x.\n",
           lapi_info.add_info);
    CALL_AND_CHECK_RC((LAPI_Init(&lapi_handle, &lapi_info)));

    /* Query LAPI for our task id */
    CALL_AND_CHECK_RC((LAPI_Qenv(lapi_handle, TASK_ID, &task_id)));

    /* Query LAPI for the total number of tasks in the job. */
    CALL_AND_CHECK_RC((LAPI_Qenv(lapi_handle, NUM_TASKS, &num_tasks)));
    printf("LAPI handle shows task_id %d and %d tasks.\n", task_id, num_tasks);

    /* both tasks will execute this before term */
    printf("Gfence on LAPI handle\n");
    CALL_AND_CHECK_RC((LAPI_Gfence(lapi_handle)));

    /* terminate the LAPI handle */
    printf("Terminate the LAPI handle\n");
    CALL_AND_CHECK_RC((LAPI_Term(lapi_handle)));

	if (!use_hndlr && udp_info)
		free(udp_info);
}
